__author__ = 'Maruf Maniruzzaman'

import tornado.ioloop
import tornado.web
import tornado.options
import tornado.template
import tornado.websocket
from cosmos.dataservice.objectservice import *
from cosmos.datamonitor.monitor import *

def start_web_service(port, endpoints, app_settings):
    application = tornado.web.Application(
        endpoints,
        **app_settings
    )

    logging.info('Starting up server on port: {0}'.format(port))
    application.listen(port)



__author__ = 'Maruf Maniruzzaman'


PASSWORD_HMAC_SIGNATURE = "hmac:"
PASSWORD_COLUMN_NAME = "password"
USER_COOKIE_NAME = "usersecret"
USER_PLAINTEXT_COOKIE_NAME = "user"
IDENTITY_TYPE_FB_GRAPH = 'fb_graph'
IDENTITY_TYPE_GOOGLE_OAUTH2 = "google_oauth2"


from cosmos.service import requesthandler

__author__ = 'Maruf Maniruzzaman'

import tornado.ioloop
import tornado.web
import tornado.options
import tornado.template
import tornado.websocket
from tornado import gen

from cosmos.service.utils import MongoObjectJSONEncoder
from cosmos.dataservice.objectservice import *

class ServiceHandler(requesthandler.RequestHandler):
    @tornado.web.asynchronous
    def get(self, object_path):
        params = object_path.split('/')
        object_name = params[0]
        id = params[1]

        db = self.settings['db']

        obj_serv = ObjectService()
        columns_str = self.get_argument("columns", None)
        filter_str = self.get_argument("filter", None)

        if filter_str:
            query=json.loads(filter_str)
        else:
            query=None

        if columns_str:
            columns = columns_str.split(',')
            columns = [column.strip() for column in columns]
        else:
            columns = []

        if id and len(id)>0:
            obj_serv.load(self.current_user, db, object_name, id, columns, self._on_load_response)
        else:
            obj_serv.find(self.current_user, db, object_name, query, columns, self._on_load_response)

    def _on_load_response(self, result, error):
        logging.debug("ServiceHandler::_on_response. Error = {0} ".format(error))
        if error:
            raise tornado.web.HTTPError(500, error)
        else:
            self.content_type = 'application/json'
            data = MongoObjectJSONEncoder().encode(result)
            self.write(data)
            self.finish()

    def _on_response(self, result, error):
        logging.debug("ServiceHandler::_on_response. Error = {0} ".format(error))
        if error:
            raise tornado.web.HTTPError(500, error)
        else:
            self.content_type = 'application/json'
            data = '{"result":"OK"}'
            self.write(data)
            self.finish()

    @tornado.web.asynchronous
    def post(self, object_path):
        params = object_path.split('/')
        object_name = params[0]

        data = json.loads(self.request.body)
        assert isinstance(data, dict)

        db = self.settings['db']

        obj_serv = ObjectService()
        obj_serv.save(self.current_user, db, object_name,data, self._on_response)


    @tornado.web.asynchronous
    def put(self, object_path):
        params = object_path.split('/')
        object_name = params[0]
        id = params[1]

        data = json.loads(self.request.body)
        assert isinstance(data, dict)

        db = self.settings['db']
        obj_serv = ObjectService()
        obj_serv.update(self.current_user, db, object_name, id, data, self._on_response)


    @tornado.web.asynchronous
    def delete(self, object_path):
        params = object_path.split('/')
        object_name = params[0]
        id = params[1]

        db = self.settings['db']
        obj_serv = ObjectService()
        obj_serv.delete(self.current_user, db, object_name, id, self._on_response)








import base64

file_data_list=[
{
'name': '/local_settings.py', 'data': base64.b64decode('aW1wb3J0IG9zCgoiIiIKWW91IHNob3VsZCBwdXQgYWxsIGRldmVsb3BtZW50IHNldHRpbmdzIGluIGhlcmUuCiIiIgoKI1RPRE86IE1ha2Ugc3VyZSB5b3UgbWF0Y2ggdGhpcyB3aXRoIHlvdXIgYW5ndWxhciBhcHBsaWNhdGlvbiBhbmQgY29weSB0aGlzIHRvIHNldHRpbmdzIGZpbGUgaW4gcHJvZHVjdGlvbgpTVEFUSUNfUEFUSCA9IG9zLnBhdGguam9pbihvcy5wYXRoLmRpcm5hbWUob3MucGF0aC5yZWFscGF0aChfX2ZpbGVfXykpLCAiYW5ndWxhci1zZWVkL2FwcCIpCklOREVYX0hUTUxfUEFUSCA9IG9zLnBhdGguam9pbihTVEFUSUNfUEFUSCwgImluZGV4Lmh0bWwiKQo=')
}]



import inspect
import datetime
import logging
import tornado.ioloop
import tornado.web
import tornado.options
import tornado.template
import tornado.websocket
from bson import ObjectId

from cosmos.rbac.service import *


__author__ = 'Maruf Maniruzzaman'

#TODO: Use context class to implement postprocessor
class CallbackContext():
    def __init__(self, *args, **kwargs):
        pass

#TODO: Use observers to check access - initialize in the __init__ of the rbac package

ACCESS_TYPE_ROLE = 1
ACCESS_TYPE_OWNER_ONLY = 2

class ObjectService():
    def __init__(self, *args, **kwargs):
        self.rbac_service = RbacService()

    def check_access(self, user, object_name, properties, access, check_owner=False):
        roles = self.rbac_service.get_roles(user)
        for role in roles:
            has_access = self.rbac_service.has_access(role, object_name, properties, access)
            if has_access:
                return ACCESS_TYPE_ROLE

            if check_owner:
                has_owner_access = self.rbac_service.has_owner_access(role, object_name, properties, access)
                if has_owner_access:
                    return ACCESS_TYPE_OWNER_ONLY

        raise tornado.web.HTTPError(401, "Unauthorized")

    def save(self, user, db, object_name, data, callback):
        logging.debug("ObjectService::save::{0}".format(object_name))
        assert isinstance(data, dict)

        properties = self.get_properties(data)
        self.check_access(user, object_name, properties, AccessType.INSERT, True)

        self.create_access_log(db, user, object_name, AccessType.INSERT)

        data['createtime'] = str(datetime.datetime.now())
        data['owner'] = str(user.get("_id"))

        preprocessor_list = get_operation_preprocessor(object_name, AccessType.INSERT)
        for preprocessor in preprocessor_list:
            preprocessor(object_name, data, AccessType.INSERT)

        db[object_name].insert(data, callback=callback)

    def find(self, user, db, object_name, query, columns, callback, limit=5000):
        logging.debug("ObjectService::find::{0}".format(object_name))
        assert inspect.ismethod(callback)

        allowed_access_type =  self.check_access(user, object_name, columns, AccessType.READ, True)

        if allowed_access_type == ACCESS_TYPE_OWNER_ONLY:
            if query:
                assert isinstance(query, dict)
                query["owner"] = str(user.get("_id"))
            else:
                query = {"owner": str(user.get("_id"))}

        self.create_access_log(db, user, object_name, AccessType.READ)

        if len(columns) > 0:
            columns_dict = {column:1 for column in columns}
        else:
            columns_dict = None

        preprocessor_list = get_operation_preprocessor(object_name, AccessType.READ)
        for preprocessor in preprocessor_list:
            preprocessor(object_name, query, AccessType.READ)

        db[object_name].find(query, columns_dict).limit(limit).to_list(limit, callback)

    def load(self, user, db, object_name, id, columns, callback):
        logging.debug("ObjectService::load::{0}".format(object_name))
        assert inspect.ismethod(callback)

        allowed_access_type = self.check_access(user, object_name, columns, AccessType.READ, True)

        if len(columns) > 0:
            columns_dict = {column:1 for column in columns}
        else:
            columns_dict = None

        query = {'_id': ObjectId(id)}

        if allowed_access_type == ACCESS_TYPE_OWNER_ONLY:
            query["owner"] = str(user.get("_id"))

        self.create_access_log(db, user, object_name, AccessType.READ)

        preprocessor_list = get_operation_preprocessor(object_name, AccessType.READ)
        for preprocessor in preprocessor_list:
            preprocessor(object_name, query, AccessType.READ)

        db[object_name].find_one(query, columns_dict, callback=callback)

    def update(self, user, db, object_name, id, data, callback):
        logging.debug("ObjectService::update::{0}".format(object_name))
        assert len(id) > 0
        assert isinstance(data, dict)
        assert inspect.ismethod(callback)

        properties = self.get_properties(data)

        allowed_access_type = self.check_access(user, object_name, properties, AccessType.UPDATE, True)

        query = {'_id': ObjectId(id)}
        if allowed_access_type == ACCESS_TYPE_OWNER_ONLY:
            query["owner"] = str(user.get("_id"))

        self.create_access_log(db, user, object_name, AccessType.UPDATE)

        data['modifytime'] = str(datetime.datetime.now())

        preprocessor_list = get_operation_preprocessor(object_name, AccessType.UPDATE)
        for preprocessor in preprocessor_list:
            preprocessor(object_name, data, AccessType.UPDATE)

        db[object_name].update(query, {'$set': data}, callback=callback)

    def delete(self, user, db, object_name, id, callback):
        logging.debug("ObjectService::delete::{0}".format(object_name))
        assert len(id) > 0
        assert inspect.ismethod(callback)

        allowed_access_type = self.check_access(user, object_name, [], AccessType.DELETE, True)

        query = {'_id': ObjectId(id)}
        if allowed_access_type == ACCESS_TYPE_OWNER_ONLY:
            query["owner"] = str(user.get("_id"))

        self.create_access_log(db, user, object_name, AccessType.DELETE)

        preprocessor_list = get_operation_preprocessor(object_name, AccessType.DELETE)
        for preprocessor in preprocessor_list:
            preprocessor(object_name, None, AccessType.DELETE)

        db[object_name].remove(query, callback=callback)

    def get_properties(self, data, namespace=None):
        properties = data.keys()
        child_props = []

        for prop in properties:
            prop_data = data.get(prop, None)

            if prop_data and isinstance(prop_data, dict):
                child_namespace = (namespace + "." + prop) if namespace else prop
                child_props = child_props + self.get_properties(prop_data, child_namespace)

        if child_props:
            properties = properties + child_props

        properties = [p if not namespace else namespace + "." + p for p in properties]

        return properties

    def create_access_log(self, db, user, module, function):
        username = None
        access_log = {"username":username, 'module': module, 'function': function}
        access_log['createtime'] = str(datetime.datetime.now())
        #db.audit.access_log.insert(access_log, callback=None)


_preprocessors = {
}

def add_operation_preprocessor(preprocessor, object_name, access_types):
    #assert inspect.ismethod(preprocessor)
    assert isinstance(access_types, collections.Iterable)

    assert len(object_name) > 0

    object_preprocessor =_preprocessors.get(object_name)
    if not object_preprocessor:
        object_preprocessor = {}
        _preprocessors[object_name] = object_preprocessor

    for access_type in access_types:
        object_acctyp_preprocessor = object_preprocessor.get(access_type)
        if not object_acctyp_preprocessor:
            object_acctyp_preprocessor = []
            object_preprocessor[access_type] = object_acctyp_preprocessor

        if not preprocessor in object_acctyp_preprocessor:
            object_acctyp_preprocessor.append(preprocessor)

def get_operation_preprocessor(object_name, access_type):
    assert len(object_name) > 0
    object_preprocessor =_preprocessors.get(object_name)

    if not object_preprocessor:
        return []

    preprocessor_list = object_preprocessor.get(access_type)
    if not preprocessor_list:
        return []

    return preprocessor_list







import collections
from mongolog import MongoHandler
import tornado
import logging
import pymongo
import cosmos
from cosmos.dataservice.objectservice import add_operation_preprocessor, ObjectService
from cosmos.rbac.object import *
from cosmos.rbac.service import update_role_cache


__author__ = 'Maruf Maniruzzaman (maruf@lekhoni.com)'


class BootLoader():
    def init_observers(self, observers):
        add_operation_preprocessor(cosmos.service.auth.before_user_insert, COSMOS_USERS_OBJECT_NAME, [AccessType.INSERT, AccessType.UPDATE])
        add_operation_preprocessor(cosmos.rbac.service.before_role_insert_update, cosmos.rbac.object.COSMOS_ROLE_OBJECT_NAME, [AccessType.INSERT, AccessType.UPDATE])

        for observer in observers:
            assert isinstance(observer, dict)
            func = observer["function"]
            object_name = observer["object_name"]
            access = observer["access"]
            assert isinstance(access, collections.Iterable)

            add_operation_preprocessor(func, object_name, access)


    def load_roles(self, db):
        object_service = ObjectService()
        object_name = COSMOS_ROLE_OBJECT_NAME
        columns = []
        object_service.find(SYSTEM_USER, db, object_name, None, columns, self._on_load_roles_response)

    def _on_load_roles_response(self, result, error):
        if error:
            raise tornado.web.HTTPError(500, error)
        else:
            for role in result:
                try:
                    update_role_cache(role)
                except ValueError, ve:
                    logging.exception("Role {0} could not be loaded.".format(role.get("name")))


    def config_mongolog(self, db_uri, db_name, log_col_name, log_level):
        c_sync = pymongo.MongoClient(db_uri, w=0)
        col = c_sync[db_name][log_col_name]
        logging.getLogger().addHandler(MongoHandler.to(collection=col))
        logging.getLogger().setLevel(log_level)


from cosmos.service import requesthandler

__author__ = 'Maruf Maniruzzaman'

from tornado import gen


class FormHandler(requesthandler.RequestHandler):
    @gen.coroutine
    def get(self):
        self.render('form.html')
